/**
* @file rmd160.c
* @brief RIPEMD-160 implementation
* @author Kosenkov
* @date 14.01.2019
*/

#define ASSERT_MODE // turn on for this file

#include "rmd160.h"
#include "memory.h"

typedef struct _RMD160BUF
{
    uint32_t A;
    uint32_t B;
    uint32_t C;
    uint32_t D;
    uint32_t E;
} RMD160BUF;

typedef struct _RMD160CTX_EX
{
    uint32_t H[5];
    uint32_t NBlocks;
    uint8_t  Buf[64];
    uint32_t Cnt;
} RMD160CTX_EX;
compiletime_assert(sizeof(RMD160CTX_EX) == sizeof(RMD160CTX));

const uint8_t g_r[10][16] =
{
    { 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15},
    { 7,  4, 13,  1, 10,  6, 15,  3, 12,  0,  9,  5,  2, 14, 11,  8},
    { 3, 10, 14,  4,  9, 15,  8,  1,  2,  7,  0,  6, 13, 11,  5, 12},
    { 1,  9, 11, 10,  0,  8, 12,  4, 13,  3,  7, 15, 14,  5,  6,  2},
    { 4,  0,  5,  9,  7, 12,  2, 10, 14,  1,  3,  8, 11,  6, 15, 13},
    { 5, 14,  7,  0,  9,  2, 11,  4, 13,  6, 15,  8,  1, 10,  3, 12},
    { 6, 11,  3,  7,  0, 13,  5, 10, 14, 15,  8, 12,  4,  9,  1,  2},
    {15,  5,  1,  3,  7, 14,  6,  9, 11,  8, 12,  2, 10,  0,  4, 13},
    { 8,  6,  4,  1,  3, 11, 15,  0,  5, 12,  2, 13,  9,  7, 10, 14},
    {12, 15, 10,  4,  1,  5,  8,  7,  6,  2, 13, 14,  0,  3,  9, 11}
};

const uint8_t g_s[10][16] =
{
    {11, 14, 15, 12,  5,  8,  7,  9, 11, 13, 14, 15,  6,  7,  9,  8},
    { 7,  6,  8, 13, 11,  9,  7, 15,  7, 12, 15,  9, 11,  7, 13, 12},
    {11, 13,  6,  7, 14,  9, 13, 15, 14,  8, 13,  6,  5, 12,  7,  5},
    {11, 12, 14, 15, 14, 15,  9,  8,  9, 14,  5,  6,  8,  6,  5, 12},
    { 9, 15,  5, 11,  6,  8, 13, 12,  5, 12, 13, 14, 11,  8,  5,  6},
    { 8,  9,  9, 11, 13, 15, 15,  5,  7,  7,  8, 11, 14, 14, 12,  6},
    { 9, 13, 15,  7, 12,  8,  9, 11,  7,  7, 12,  7,  6, 15, 13, 11},
    { 9,  7, 15, 11,  8,  6,  6, 14, 12, 13,  5, 14, 13, 13,  7,  5},
    {15,  5,  8, 11, 14, 14,  6, 14,  6,  9, 12,  9, 12,  5, 15,  8},
    { 8,  5, 12,  9, 12,  5, 14,  6,  8, 13,  6,  5, 15, 13, 11, 11}
};

const uint32_t g_KL[5] = {0x00000000, 0x5A827999, 0x6ED9EBA1, 0x8F1BBCDC, 0xA953FD4E};
const uint32_t g_KR[5] = {0x50A28BE6, 0x5C4DD124, 0x6D703EF3, 0x7A6D76E9, 0x00000000};

typedef uint32_t (*RMD160F)(uint32_t, uint32_t, uint32_t);
uint32_t RMD160F0(uint32_t x, uint32_t y, uint32_t z) { return x ^ y ^ z; }
uint32_t RMD160F1(uint32_t x, uint32_t y, uint32_t z) { return (x & y) | (~x & z); }
uint32_t RMD160F2(uint32_t x, uint32_t y, uint32_t z) { return (x | ~y) ^ z; }
uint32_t RMD160F3(uint32_t x, uint32_t y, uint32_t z) { return (x & z) | (y & ~z); }
uint32_t RMD160F4(uint32_t x, uint32_t y, uint32_t z) { return x ^ (y | ~z); }
const RMD160F g_RMD160Fs[5] = {RMD160F0, RMD160F1, RMD160F2, RMD160F3, RMD160F4};

uint32_t rol(uint32_t x, uint8_t n) { return (x << n) | (x >> (32 - n)); }

void R(RMD160BUF* buf, RMD160F f, uint32_t k, uint32_t r, uint32_t s, uint32_t* x)
{
    uint32_t tmp = rol(buf->A + f(buf->B, buf->C, buf->D) + k + x[r], s) + buf->E;
    buf->A = buf->E;
    buf->E = buf->D;
    buf->D = rol(buf->C, 10);
    buf->C = buf->B;
    buf->B = tmp;
}
                       
void rmd160block(RMD160CTX_EX* ctx, const uint8_t* data)
{
    uint8_t i;
    uint8_t j;
    RMD160BUF bufL;
    RMD160BUF bufR;
    uint32_t tmp;
    uint32_t* x = (uint32_t*)data;

    mldrMemCpy((void*)&bufL, ctx->H, sizeof(bufL)); /* left lane */
    mldrMemCpy((void*)&bufR, ctx->H, sizeof(bufR)); /* right lane */
    
    for (i = 0; i < 5; ++i)
    {
        for (j = 0; j < 16; ++j)
        {
            R(&bufL, g_RMD160Fs[    i], g_KL[i], g_r[    i][j], g_s[    i][j], x);
            R(&bufR, g_RMD160Fs[4 - i], g_KR[i], g_r[5 + i][j], g_s[5 + i][j], x);
        }
    }

    tmp       = ctx->H[1] + bufL.C + bufR.D;
    ctx->H[1] = ctx->H[2] + bufL.D + bufR.E;
    ctx->H[2] = ctx->H[3] + bufL.E + bufR.A;
    ctx->H[3] = ctx->H[4] + bufL.A + bufR.B;
    ctx->H[4] = ctx->H[0] + bufL.B + bufR.C;
    ctx->H[0] = tmp;
}

void rmd160Init(RMD160CTX* ctx)
{
    RMD160CTX_EX* ctxEx = (RMD160CTX_EX*)ctx;
    ctxEx->H[0] = 0x67452301;
    ctxEx->H[1] = 0xEFCDAB89;
    ctxEx->H[2] = 0x98BADCFE;
    ctxEx->H[3] = 0x10325476;
    ctxEx->H[4] = 0xC3D2E1F0;
    ctxEx->NBlocks = 0;
    ctxEx->Cnt = 0;
}

void rmd160Update(RMD160CTX* ctx, const uint8_t* data, Sz8 sz)
{
    RMD160CTX_EX* ctxEx = (RMD160CTX_EX*)ctx;
    if (ctxEx->Cnt == 64) // flush the buffer
    {
        rmd160block(ctxEx, ctxEx->Buf);
        ctxEx->Cnt = 0;
        ctxEx->NBlocks++;
    }
    if (!data || !sz)
    {
        return;
    }
    if (ctxEx->Cnt)
    {
        for (; sz && (ctxEx->Cnt < 64); --sz)
        {
            ctxEx->Buf[ctxEx->Cnt++] = *(data++);
        }
        rmd160Update(ctx, NULL, 0); // calc and return
    }

    while (sz >= 64)
    {
        rmd160block(ctxEx, data);
        ctxEx->NBlocks++;
        sz -= 64;
        data += 64;
    }
    for (; sz && (ctxEx->Cnt) < 64; --sz)
    {
        ctxEx->Buf[ctxEx->Cnt++] = *(data++);
    }
}

uint8_t* rmd160Final(RMD160CTX* ctx, const uint8_t* data, Sz8 sz)
{
    uint32_t t;
    uint32_t msb;
    uint32_t lsb;
    RMD160CTX_EX* ctxEx = (RMD160CTX_EX*)ctx;

    rmd160Update(ctx, data, sz);
    
    t = ctxEx->NBlocks;
    /* multiply by 64 to make a byte count */
    lsb = t << 6;
    msb = t >> 26;
    /* add the count */
    t = lsb;
    if ((lsb += ctxEx->Cnt ) < t)
    {
        ++msb;
    }
    /* multiply by 8 to make a bit count */
    t = lsb;
    lsb <<= 3;
    msb <<= 3;
    msb |= t >> 29;

    if (ctxEx->Cnt < 56) // enough room
    {
        ctxEx->Buf[ctxEx->Cnt++] = 0x80;    // pad
        mldrMemSet(&ctxEx->Buf[ctxEx->Cnt], 0, 56 - ctxEx->Cnt);
        ctxEx->Cnt = 56;
    }
    else // need one extra block
    {
        ctxEx->Buf[ctxEx->Cnt++] = 0x80;    // pad
        mldrMemSet(&ctxEx->Buf[ctxEx->Cnt], 0, 64 - ctxEx->Cnt);
        ctxEx->Cnt = 64;
        rmd160Update(ctx, NULL, 0);
        mldrMemSet(ctxEx->Buf, 0, 56);      // fill next block with zeroes
    }
    /* append the 64 bit count */
    ctxEx->Buf[56] = lsb;
    ctxEx->Buf[57] = lsb >> 8;
    ctxEx->Buf[58] = lsb >> 16;
    ctxEx->Buf[59] = lsb >> 24;
    ctxEx->Buf[60] = msb;
    ctxEx->Buf[61] = msb >> 8;
    ctxEx->Buf[62] = msb >> 16;
    ctxEx->Buf[63] = msb >> 24;
    rmd160block(ctxEx, ctxEx->Buf);

    return (uint8_t*)ctxEx->H;
}

