#ifndef __CRYPTO_ALU_H__
#define __CRYPTO_ALU_H__

#include "memory.h"

#define MAX_SZ 0xFFFFFFFF
#define aluGetSz(sz) ((sz + 3) >> 2)

typedef uint8_t Reg128Num;

/**
* ALU initialization
*
* @author Kosenkov
* @date 18.05.2018
*/
void aluInitCCPU(void);

/**
* Writes 32-bits word to ALU register
*
* @param num Number of 128-bits ALU register
* @param offs 32-bits offset in 128-bits register
* @param data Word to write
* @author Kosenkov
* @date 18.05.2018
*/
void aluWriteWord(Reg128Num num, Sz32 offs, uint32_t data);

/**
* Reads 32-bits word from ALU register
*
* @param num Number of 128-bits ALU register
* @param offs 32-bits offset in 128-bits register
* @param data Word to read to
* @author Kosenkov
* @date 18.05.2018
*/
void aluReadWord(Reg128Num num, Sz32 offs, uint32_t* data);

/**
* Writes array of 32-bits words to ALU register
*
* @param num Number of 128-bits ALU register
* @param sz Size of array in 32-bits words
* @param data Array to write
* @author Kosenkov
* @date 18.05.2018
*/
void aluWriteData(Reg128Num num, Sz32 sz, uint32_t* data);

/**
* Reads array of 32-bits words from ALU register
*
* @param num Number of 128-bits ALU register
* @param sz Size of array in 32-bits words
* @param data Array to read to
* @author Kosenkov
* @date 18.05.2018
*/
void aluReadData(Reg128Num num, Sz32 sz, uint32_t* data);

/**
* Returns carry flag from ALU
*
* @return CARRY flag  if carry bit was set, 0 othrwise
* @author Kosenkov
* @date 18.05.2018
*/
uint8_t aluGetCarry(void);

/**
* Returns shift flag from ALU
*
* @return SHIFT flag if shift bit was set, 0 othrwise
* @author Kosenkov
* @date 18.05.2018
*/
uint8_t aluGetShift(void);

/**
* Zeroes 128-bits registers
*
* @param res Register to be zeroed
* @param sz Count of registers to be zeroed
* @author Kosenkov
* @date 18.05.2018
*/
void aluZero(Reg128Num res, Sz128 sz);

/**
* Copy data from register a to register res
*
* @param res Resulting register
* @param a 1st operand
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluMov(Reg128Num res, Reg128Num a, Sz128 sz);

/**
* Move with carry flag usage
*
* @param res Resulting register
* @param a 1st operand
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluCMov(Reg128Num res, Reg128Num a, Sz128 sz);

/**
* res = a + b
*
* @param res Resulting register
* @param a 1st operand
* @param b 2st operand
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluAdd(Reg128Num res, Reg128Num a, Reg128Num b, Sz128 sz);

/**
* res = a - b
*
* @param res Resulting register
* @param a 1st operand
* @param b 2st operand
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluSub(Reg128Num res, Reg128Num a, Reg128Num b, Sz128 sz);

/**
* res = a << 1
*
* @param res Resulting register
* @param a 1st operand
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluShl(Reg128Num res, Reg128Num a, Sz128 sz);

/**
* res = a << 1 + carry (shift with carry)
*
* @param res Resulting register
* @param a 1st operand
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluCShl(Reg128Num res, Reg128Num a, Sz128 sz);

/**
* res = a >> 1
*
* @param res Resulting register
* @param a 1st operand
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluShr(Reg128Num res, Reg128Num a, Sz128 sz);

/**
* res = a >> 1 + carry (shift with carry)
*
* @param res Resulting register
* @param a 1st operand
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluCShr(Reg128Num res, Reg128Num a, Sz128 sz);

/**
* res = a / b
* a = a % b
*
* @param res Resulting register
* @param a 1st operand
* @param b 2st operand
* @param tmp Temporary register
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluDiv(Reg128Num res, Reg128Num a, Reg128Num b, Reg128Num tmp, Sz128 sz);

/**
* a = a % b
*
* @param a 1st operand
* @param b 2st operand
* @param tmp Temporary register
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluMod(Reg128Num a, Reg128Num b, Reg128Num tmp, Sz128 sz);

/**
* Exchanges data of registers
*
* @param left 1st operand
* @param right 2st operand
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluXchg(Reg128Num left, Reg128Num right, Sz128 sz);

/**
* res = a * b. res is doublesized
*
* @param res Resulting register (its size is sz * 2)
* @param a 1st operand
* @param b 2st operand
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluMul(Reg128Num res, Reg128Num a, Reg128Num b, Sz128 sz);

/**
* res = [least significant half of] a * b. res is singlesized
*
* @param res Resulting register
* @param a 1st operand
* @param b 2st operand
* @param sz Count of registers to operate
* @author Kosenkov
* @date 18.05.2018
*/
void aluHalfMul(Reg128Num res, Reg128Num a, Reg128Num b, Sz128 sz);

#endif
