/**
* @file gpio.c
* @brief GPIO unit implementation
* @author Kosenkov
* @date 29.03.2018
*/

#include "cm0.h"
#include "types.h"

////////////////////////////////////////////////////////////////////////////////
// GPIO helpers
////////////////////////////////////////////////////////////////////////////////
/**
* Gets pin 2-bits param
*
* @param pin Number
* @param reg Register address
* @return 2-bits param
* @author Kosenkov
* @date 18.05.2018
*/
uint8_t get2BitsParam(uint8_t pin, __IO uint32_t* reg)
{
    return ((*reg >> (pin << 1)) & 3);
}

/**
* Sets pin 2-bits param
*
* @param mode 2-bits param to set
* @param pin Number
* @param reg Register address
* @author Kosenkov
* @date 18.05.2018
*/
void set2BitsParam(uint8_t mode, uint8_t pin, __IO uint32_t* reg)
{
    uint32_t mask = ~((uint32_t)3 << (pin << 1));
    uint32_t target = (uint32_t)mode << (pin << 1);
    *reg = ((*reg & mask) | target);
}

////////////////////////////////////////////////////////////////////////////////
// GPIO implementation
////////////////////////////////////////////////////////////////////////////////
/**
* Gets pin mode
*
* @param pin Number
* @return Mode
* @author Kosenkov
* @date 18.05.2018
*/
uint8_t gpioGetPinMode(uint8_t pin)
{
    return get2BitsParam(pin, &GPIO->PORT_MODE);
}

/**
* Sets pin mode
*
* @param mode To set
* @param pin Number
* @author Kosenkov
* @date 18.05.2018
*/
void gpioSetPinMode(uint8_t mode, uint8_t pin)
{
    set2BitsParam(mode, pin, &GPIO->PORT_MODE);
}

/**
* Gets pin slew rate
*
* @param pin Number
* @return Slew rate
* @author Kosenkov
* @date 18.05.2018
*/
uint8_t gpioGetSlewRate(uint8_t pin)
{
    return get2BitsParam(pin, &GPIO->SLEW_RATE);
}

/**
* Sets pin slew rate
*
* @param sr Slew rate to set
* @param pin Number
* @author Kosenkov
* @date 18.05.2018
*/
void gpioSetSlewRate(uint8_t mode, uint8_t pin)
{
    set2BitsParam(mode, pin, &GPIO->SLEW_RATE);
}

/**
* Gets pin pull mode (pull up/pull down)
*
* @param pin Number
* @return Pull mode
* @author Kosenkov
* @date 18.05.2018
*/
uint8_t gpioGetPUPD(uint8_t pin)
{
    return get2BitsParam(pin, &GPIO->PU_PD);
}

/**
* Sets pin pull mode (pull up/pull down)
*
* @param pupd Mode to set
* @param pin Number
* @author Kosenkov
* @date 18.05.2018
*/
void gpioSetPUPD(uint8_t mode, uint8_t pin)
{
    set2BitsParam(mode, pin, &GPIO->PU_PD);
}
