/**
* @file noise.c
* @brief Noise generator implementation
* @author Kosenkov
* @date 18.04.2018
*/

#include "noise.h"
#include "rng.h"
#include "cm0.h"

#include "rngTest.h"

/**
* Noise module initialization
*
* @param output Random output mode on/off
* @param cascade Cascade mode on/off
* @retval CM0_OK if success
* @retval CM0_NOISE_ERR otherwise
* @author Kosenkov
* @date 18.05.2018
*/
CM0_STATUS noiseInit(uint8_t cascade, uint8_t output)
{
    uint32_t i;
    uint32_t seeds[NOISE_RAND_SZ];
    uint32_t modeMask = NOISE_MODE_0 | NOISE_MODE_1 | NOISE_MODE_2 | NOISE_MODE_3;
    uint32_t cascadeMask = NOISE_CASCADE_1 | NOISE_CASCADE_2 | NOISE_CASCADE_3;
    uint32_t outputMask = NOISE_ALLOW_OUTPUT_0 | NOISE_ALLOW_OUTPUT_1 | NOISE_ALLOW_OUTPUT_2 | NOISE_ALLOW_OUTPUT_3;
    
    // get 4x32 bits randoms from RNG
    rngSetMode(RNG_WORK_MODE_AFTER_READ);
    for (i = 0; i < 4; ++i)
    {
        if (CM0_CHK_ERR(rngGetWord(&seeds[i])))
        {
            rngSetMode(RNG_WORK_MODE_STOPPED);
            return CM0_NOISE_ERR;
        }            
    }
    rngSetMode(RNG_WORK_MODE_STOPPED);
    
    NOISE_GEN->CTRL &= (~modeMask); // init state
    for (i = 0; i < 4; ++i)
    {
        NOISE_GEN->SEED[i] = seeds[i];
    }
    if (cascade)
    {
        NOISE_GEN->CTRL |= cascadeMask;
    }
    else
    {
        NOISE_GEN->CTRL &= (~cascadeMask);
    }
    if (output)
    {
        NOISE_GEN->CTRL |= outputMask;
    }
    else
    {
        NOISE_GEN->CTRL &= (~outputMask);
    }
    NOISE_GEN->CTRL |= modeMask; // work state
    
    return CM0_OK;
}

/**
* Get random number from noise module
*
* @param rnd Array for randoms
* @param sz Size of array
* @author Kosenkov
* @date 18.05.2018
*/
void noiseRandom(uint32_t* rnd, Sz32 sz)
{
    uint8_t i;
    sz = sz > 4 ? 4 : sz;
        
    for (i = 0; i < sz; ++i)
    {
        rnd[i] = NOISE_GEN->RAND[i];
    }
}

/**
* Runs random quality tests using noise generator
* 
* @retval CM0_OK if success
* @retval CM0_RNG_TEST_ERR otherwise
* @author Kosenkov
* @date 25.05.2018
*/
CM0_STATUS noiseTest()
{
    const Sz32 sz = STAT_CNT / sizeof(uint32_t);
    const Sz32 noiseSz = 4;
    uint8_t random[noiseSz * sizeof(uint32_t)];
    uint8_t i;
    uint16_t j;
    uint8_t k;
    
    // init statistical tests
    rngTestInit();
    
    for (i = 0; i < 3; ++i)
    {
        // get randoms from noise module for testing
        for (j = 0; j < sz; j += noiseSz)
        {
            // gen randoms
            noiseRandom((uint32_t*)random, noiseSz);
            
            // test randoms
            for (k = 0; k < noiseSz * sizeof(uint32_t); ++k)
            {
                rngTestByte(random[k]);
            }
        }
        
        // check results
        if (CM0_CHK_ERR(rngTestResult()))
        {
            continue;
        }
        else
        {
            return CM0_OK;
        }
    }
    
    return CM0_RNG_TEST_ERR;
}
