/**
* @file uart.c
* @brief UART implementation
* @author Kosenkov
* @date 15.02.2018
*/

#include "cm0.h"
#include "crc32.h"
#include "gpio.h"

#define RX_PIN ((uint8_t)6)
#define TX_PIN ((uint8_t)5)

typedef struct
{
    uint8_t RxMode;
    uint8_t TxMode;
    uint8_t RxSlewRate;
    uint8_t TxSlewRate;
    uint8_t RxPUPD;
    uint8_t TxPUPD;
} SavedData;

SavedData gSD __attribute__((section(".bss")));
uint32_t gCrc __attribute__((section(".bss")));

/**
* UART module initialization
*
* @param clock Current clock on UART
* @param baudRate Desired baud rate
* @author Kosenkov
* @date 18.05.2018
*/
void uartInit(uint32_t clock, uint32_t baudRate)
{
    // save initial state
    gSD.RxMode = gpioGetPinMode(RX_PIN);
    gSD.TxMode = gpioGetPinMode(TX_PIN);
    gSD.RxSlewRate = gpioGetSlewRate(RX_PIN);
    gSD.TxSlewRate = gpioGetSlewRate(TX_PIN);
    gSD.RxPUPD = gpioGetPUPD(RX_PIN);
    gSD.TxPUPD = gpioGetPUPD(TX_PIN);
    gCrc = crc32(CRC32_INITIAL_CTX, (uint8_t*)&gSD, sizeof(gSD));
    
    // set GPIO pins to UART configuration
    gpioSetPinMode(GPIO_MODE_AF, RX_PIN);
    gpioSetPinMode(GPIO_MODE_AF, TX_PIN);
    gpioSetSlewRate(GPIO_SLEWRATE_HIGH, RX_PIN);
    gpioSetSlewRate(GPIO_SLEWRATE_HIGH, TX_PIN);
    gpioSetPUPD(GPIO_PUPD_NONE, RX_PIN);
    gpioSetPUPD(GPIO_PUPD_NONE, TX_PIN);
    
    UART->BRR = clock/baudRate;
    UART->CR1 = UART_CR1_RE | UART_CR1_TE | UART_CR1_UE;
}

/**
* UART module deinitialization
*
* @author Kosenkov
* @date 23.05.2018
*/
void uartDeinit()
{
    if (gCrc != crc32(CRC32_INITIAL_CTX, (uint8_t*)&gSD, sizeof(gSD)))
    {
        return;
    }

    // set GPIO pins to initial state
    gpioSetPinMode(gSD.RxMode, RX_PIN);
    gpioSetPinMode(gSD.TxMode, TX_PIN);
    gpioSetSlewRate(gSD.RxSlewRate, RX_PIN);
    gpioSetSlewRate(gSD.TxSlewRate, TX_PIN);
    gpioSetPUPD(gSD.RxPUPD, RX_PIN);
    gpioSetPUPD(gSD.TxPUPD, TX_PIN);
    
    UART->BRR = 0;
    UART->CR1 = 0;
}

/**
* Write to UART. Blocking call
*
* @param data Data for writing to UART
* @param sz Size of data
* @author Kosenkov
* @date 18.05.2018
*/
void uartWrite(uint8_t* data, uint32_t sz)
{
    uint32_t i;
    for (i = 0; i < sz; ++i)
    {
        while (!(UART->SR & UART_SR_TXE));
        UART->DR = data[i];
        while (!(UART->SR & UART_SR_TC));
    }
}

/**
* Reads from UART. Blocking call
*
* @param data Data for reading from UART
* @param sz Size of data
* @author Kosenkov
* @date 18.05.2018
*/
void uartRead(uint8_t* data, uint32_t sz)
{
    uint32_t i;
    for (i = 0; i < sz; ++i)
    {
        while (!(UART->SR & UART_SR_RXNE));
        data[i] = (uint8_t)UART->DR;
    }
}
